# coding=utf-8
from libopensesame.py3compat import *
import js2py
from js2py.translators import translate_js
from libopensesame.inline_script import inline_script
from libqtopensesame.items.inline_script import inline_script as qtinline_script
from libqtopensesame.items.qtplugin import qtplugin
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'inline_javascript', category=u'plugin')


class JavaScriptWorkspace(js2py.EvalJs):

	def __init__(self, experiment):

		self.__dict__[u'experiment'] = experiment
		self.__dict__[u'_globals'] = {}
		self.init_globals()
		js2py.EvalJs.__init__(self, self._globals)

	def init_globals(self):

		"""
		desc:
			Initialize the global workspace.
		"""

		# By setting the __name__ global, the workspace will operate as a
		# module, so that e.g. import statements don't suffer from locality.
		self._globals.update({
			u'__name__'			: u'javascript_workspace',
			u'vars'				: self.experiment.var,
		})

	def _compile(self, js=None, use_compilation_plan=False):

		code = translate_js(
			js,
			'',
			use_compilation_plan=use_compilation_plan
		)
		return compile(
			code,
			'<EvalJS snippet>',
			'exec'
		)

	def _exec(self, compiled):

		exec(compiled, self._context)


class inline_javascript(inline_script):

	description = u'Executes JavaScript code (ECMA 5.1)'

	def reset(self):

		self.var._prepare = u''
		self.var._run = u''

	@property
	def workspace(self):

		if not hasattr(self.experiment, u'javascript_workspace'):
			self.experiment.javascript_workspace = JavaScriptWorkspace(
				self.experiment
			)
		return self.experiment.javascript_workspace


class qtinline_javascript(qtinline_script):

	description = _(u'Executes JavaScript code (ECMA 5.1)')
	language = u'JavaScript'  # For OpenSesame 3.2
	ext = u'.js'  # For OpenSesame 3.3
	mime_type = u'application/javascript'  # For OpenSesame 3.3

	def __init__(self, name, experiment, string=None):

		inline_script.__init__(self, name, experiment, string)
		qtplugin.__init__(self, plugin_file=__file__)

	def item_icon(self):

		return self.qicon



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
import os
import tempfile
import webbrowser
from qtpy.QtWidgets import QFileDialog
from qtpy.QtCore import QRegExp
from qtpy.QtGui import QRegExpValidator, QIcon
from libqtopensesame.widgets.base_widget import base_widget
from libopensesame.osexpfile import osexpwriter
from osweb import export, __version__
from libqtopensesame.misc.translate import translation_context
from libqtopensesame.misc.config import cfg
_ = translation_context(u'oswebext', category=u'extension')

MEGABYTE = 1024 ** 2


class oswebext_widget(base_widget):

	"""
	desc:
		Controls for interacting with OSWeb
	"""

	def __init__(self, main_window, oswebext):

		"""
		desc:
			Constructor.

		arguments:
			main_window:	The main-window object.
		"""

		super(oswebext_widget, self).__init__(
			main_window,
			ui=u'extensions.oswebext.oswebext'
		)
		self._oswebext = oswebext
		self.ui.button_test.clicked.connect(self._test)
		self.ui.button_jatos.clicked.connect(self._export_jatos)
		self.ui.button_convert.clicked.connect(self._convert_results)
		self.ui.label_version.setText(__version__)
		self.ui.linedit_subject.setValidator(
			QRegExpValidator(QRegExp("^(?:\d+(?:-\d+)?(?:,(?!$))?)+"))
		)
		self.ui.icon_expsize_warning.setPixmap(
			QIcon.fromTheme('emblem-important').pixmap(32, 32)
		)
		self._run_linter()

	def on_activate(self):

		self._run_linter()
		self._check_filesize()

	def set_error(self, error_report):

		self.ui.label_linter.setText(error_report)

	def _run_linter(self):

		self._oswebext.run_linter()

	def _check_filesize(self):

		try:
			size = self.pool.size()
		except:
			size = -1
		if size > 10 * MEGABYTE:
			self.ui.label_expsize_warning.setText(_(
				u'Your experiment is %d MB. <br />'
				u'This exceeds the recommended maximum size of 10 MB. <br /> '
				u'This may increase online loading time.'
			) % (size // MEGABYTE))
			self.ui.icon_expsize_warning.setVisible(True)
			self.ui.label_expsize_warning.setVisible(True)
		else:
			self.ui.icon_expsize_warning.setVisible(False)
			self.ui.label_expsize_warning.setVisible(False)

	def _test(self):

		"""
		desc:
			Tests the experiment by running it in an external browser.
		"""

		osexp = self._tmp_osexp()
		html = self._tmp_html()
		poss_subject_nrs = self.ui.linedit_subject.text()
		fullscreen = self.ui.fs_checkBox.isChecked()
		export.standalone(
			osexp,
			html,
			subject=poss_subject_nrs,
			fullscreen=fullscreen
		)
		webbrowser.open('file://{}'.format(html))
		os.remove(osexp)

	def _export_jatos(self):

		if self.main_window.current_path:
			suggested_path = self.main_window.current_path + u'.zip'
		else:
			suggested_path = cfg.file_dialog_path
		path = QFileDialog.getSaveFileName(
			self.main_window,
			_(u'Export JATOS study…'),
			directory=suggested_path,
			filter=u'JATOS study (*.zip)'
		)
		if isinstance(path, tuple):
			path = path[0]
		if not path:
			return
		osexp = self._tmp_osexp()
		poss_subject_nrs = self.ui.linedit_subject.text()
		fullscreen = self.ui.fs_checkBox.isChecked()
		export.jatos(
			osexp,
			path,
			title=self.experiment.title,
			description=self.experiment.description,
			subject=poss_subject_nrs,
			fullscreen=fullscreen
		)
		os.remove(osexp)
		self.extension_manager.fire(
			'notify',
			message='Experiment succesfully exported',
			category='success',
			always_show=True
		)

	def _convert_results(self):

		from osweb import data
		from datamatrix import io

		jatos_results_path = QFileDialog.getOpenFileName(
			self.main_window,
			_(u'Select JATOS results file…'),
			filter=u'JATOS results (*.txt)'
		)
		if isinstance(jatos_results_path, tuple):
			jatos_results_path = jatos_results_path[0]
		if not jatos_results_path:
			return

		self.main_window.set_busy(True)
		try:
			dm = data.parse_jatos_results(jatos_results_path)
		finally:
			self.main_window.set_busy(False)
		export_path = QFileDialog.getSaveFileName(
			self.main_window,
			_(u'Save as…'),
			filter=u'Excel (*.xlsx);;CSV (*.csv)'
		)
		if isinstance(export_path, tuple):
			export_path = export_path[0]
		if not export_path:
			return
		if export_path.lower().endswith(u'.xlsx'):
			io.writexlsx(dm, export_path)
		else:
			io.writetxt(dm, export_path)

	def _tmp_osexp(self):

		with tempfile.NamedTemporaryFile(suffix='.osexp', delete=False) as fd:
			pass
		osexpwriter(self.experiment, fd.name)
		return fd.name

	def _tmp_html(self):

		with tempfile.NamedTemporaryFile(suffix='.html', delete=False) as fd:
			pass
		return fd.name



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

import json
from datamatrix import DataMatrix
from datamatrix.py3compat import *


STEPS = 100


def parse_jatos_results(jatos_path):

	if hasattr(json.decoder, 'JSONDecodeError'):
		jsonerror = json.decoder.JSONDecodeError
	else:
		jsonerror = ValueError
	dm = DataMatrix(length=STEPS)
	invalid_lines = 0
	incomplete_lines = 0
	total_lines = 0
	row = 0
	with open(jatos_path) as fd:
		for line in fd:
			# Strip the lines so that they are valid json
			line = line.strip()
			if line.startswith('[{'):
				line = line[1:]
			if line.endswith('},') or line.endswith('}]'):
				line = line[:-1]
			total_lines += 1
			# Ignore empty lines
			if not line:
				continue
			try:
				d = json.loads(line)
			except jsonerror:
				invalid_lines += 1
				continue
			else:
				if not d:
					continue
			# Complete data is stored as a single-json object with a data and
			# context field. We're only interested in the data field here,
			# which in turn is a list of dicts where each dict corresponds to
			# a logger call. This data is stored when an experiment is
			# successfully finished.
			if len(d) == 2 and 'data' in d and 'context' in d:
				trials = d['data']
			# Incomplete data is stored as a single json line corresponding to
			# a single logger call. This kind of data should only happen when
			# the experiment is not finished, so that data is only partially
			# logged, one trial at a time.
			else:
				trials = [d]
				incomplete_lines += 1
			for trial in trials:
				if row >= len(dm):
					dm.length += STEPS
				for key, val in trial.items():
					if key not in dm:
						dm[key] = u''
					dm[key][row] = safe_decode(val)
				row += 1
	dm.length = row
	if invalid_lines:
		warn('Failed to parse {} of {} lines'.format(
			invalid_lines,
			total_lines
		))
	if incomplete_lines:
		warn('Incomplete (unfinished) data in {} of {} lines'.format(
			incomplete_lines,
			total_lines
		))
	return dm


if __name__ == '__main__':

	import sys
	dm = parse_jatos_results(sys.argv[-1])
	print(dm)
	print('{} trials'.format(len(dm)))



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

SUPPORTED_ITEMS = [
	u'sequence',
	u'loop',
	u'sketchpad',
	u'feedback',
	u'mouse_response',
	u'sampler',
	u'keyboard_response',
	u'logger',
	u'touch_response',
	u'notepad',
	u'reset_feedback',
	u'repeat_cycle',
	u'advanced_delay',
	u'inline_javascript'
]


def check_compatibility(exp):

	return u'\n'.join([
		u'Item {} is not supported'.format(item.item_type)
		for item in exp.items.values()
		if item.item_type not in SUPPORTED_ITEMS
	])



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from libqtopensesame.extensions import base_extension
from libqtopensesame.misc.translate import translation_context
from osweb import linter
_ = translation_context(u'oswebext', category=u'extension')


class oswebext(base_extension):

	"""
	desc:
		An example extension that lists all available events.
	"""

	def event_startup(self):

		self._widget = None
		self._show_controls_action = self.qaction(
			u'applications-internet',
			u'OSWeb',
			self._show_controls,
		)
		self.add_action(
			self.get_submenu(u'tools'),
			self._show_controls_action,
			3,
			False,
			False
		)

	def activate(self):

		self.widget()._test()

	def event_new_item(self, name, _type):

		self.run_linter()

	def event_delete_item(self, name):

		self.run_linter()

	def event_purge_unused_items(self):

		self.run_linter()

	def run_linter(self):

		error_report = linter.check_compatibility(self.experiment)
		if not error_report:
			error_report = u'No problems detected'
			self.action.setEnabled(True)
		else:
			self.action.setEnabled(False)
		if self._widget is not None:
			self._widget.set_error(error_report)

	def _show_controls(self):

		"""
		desc:
			Open the plug-in manager tab, or switch to it if already open.
		"""

		self.tabwidget.add(self.widget(), self.icon(), self.label())

	def widget(self):

		"""
		returns:
			desc:	The widget to be used in the tab widget.
			type:	QWidget
		"""

		if self._widget is None:
			self.set_busy()
			from oswebext_widget import oswebext_widget
			self._widget = oswebext_widget(self.main_window, self)
			self.set_busy(False)
		return self._widget



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

import os
import hashlib
import time
import json
import base64
import tempfile
import zipfile
import io
import sys
from bs4 import BeautifulSoup

src_folder = os.path.join(os.path.dirname(__file__), u'src')
# Folder containing osweb javascript
srcPaths = {
    'js': os.path.join(src_folder, u'js'),
    'css': os.path.join(src_folder, u'css'),
    'html': os.path.join(src_folder, u'html'),
    'img': os.path.join(src_folder, u'img')
}
py3 = sys.version_info[0] >= 3


def safe_decode(s):

    if (py3 and isinstance(s, bytes)) or (not py3 and isinstance(s, str)):
        return s.decode('utf-8')
    return s


def standalone(osexp, dst, subject='0', fullscreen=False):

    params = {'subject': subject, 'fullscreen': fullscreen}
    _html(osexp, dst, u'standalone', params)


def jatos(
        osexp,
        dst,
        title='My OpenSesame experiment',
        description='No description',
        subject='0',
        fullscreen=False
):

    uuid = _unique_hash()
    dirname = tempfile.mkdtemp(suffix=u'.jatos')
    os.mkdir(os.path.join(dirname, uuid))
    index_path = os.path.join(dirname, uuid, u'index.html')
    jas_path = os.path.join(dirname, u'info.jas')

    params = {
        'subject': subject,
        'fullscreen': fullscreen
    }

    assets = _html(
        osexp,
        index_path,
        u'jatos',
        params
    )

    info = {
        'version': '3',
        'data': {
            'title': title,
            'description': description,
            'groupStudy': False,
            'dirName': uuid,
            'comments': 'Experiment exported by OpenSesame',
            'jsonData': None,
            'componentList': [{
                'title': 'OSWeb experiment',
                'htmlFilePath': 'index.html',
                'reloadable': False,
                'active': True,
                'comments': """The following variables can be passed to the OSWeb runner with the JSON Input field:
- subject (string): the allowed range of subject numbers. If empty or ommitted, the current jatos component result id will be used for the subject number.
The value can be a range ('1-10'), a comma-separated list of numbers ('1,2,3,6,7,8') or a combination of the two ('1,2,3,5-10,25')
- fullscreen (bool): indicates whether the experiment should be run fullscreen
- omitJatosIds (bool): If set to true, Jatos specific data will not be appended to the results
				""",
                'jsonData': json.dumps(params),
            }],
            'batchList': []
        }
    }

    with io.open(jas_path, u'w', encoding=u'utf-8') as fd:
        fd.write(safe_decode(json.dumps(info)))

    with zipfile.ZipFile(dst, 'w') as fd:
        fd.write(jas_path, u'info.jas')
        fd.write(index_path, os.path.join(uuid, u'index.html'))
        for img in [u'opensesame.png', u'warning.png']:
            fd.write(os.path.join(srcPaths['img'], img),
                     os.path.join(uuid, u'img', img))
        fd.write(osexp, os.path.join(uuid, os.path.basename(osexp)))
        for js in assets['js']:
            fd.write(js['src'], os.path.join(uuid, js['dest']))
        for css in assets['css']:
            fd.write(css['src'], os.path.join(uuid, css['dest']))


# Private functions

def _get_os_assets(type):

    return [{
            'src': os.path.join(srcPaths[type], basename),
            'dest': type + '/' + basename
            } for basename in os.listdir(srcPaths[type])
            if basename.startswith((u'osweb', u'vendors~osweb'))
            ]


def _html(osexp, dst, mode, params=None):

    assets = {
        'js': _get_os_assets(u'js'),
        'css': _get_os_assets(u'css')
    }

    # Add potentially present platform js file.
    env_js = u'{}.js'.format(mode)
    env_js_src = os.path.join(srcPaths[u'js'], env_js)
    if os.path.exists(env_js_src):
        assets['js'].append({
            'src': env_js_src,
            'dest': (u'js/' + env_js)
        })

    # Add potentially present platform css file.
    # Not very DRY code....
    env_css = u'{}.css'.format(mode)
    env_css_src = os.path.join(srcPaths[u'css'], env_css)
    if os.path.exists(env_css_src):
        assets['css'].append({
            'src': env_css_src,
            'dest': (u'css/' + env_css)
        })

    tmpl = os.path.join(srcPaths['html'], u'{}.html'.format(mode))

    # The HTML file parsed as a DOM Tree
    with open(tmpl, 'r') as t_fp:
        dom = BeautifulSoup(t_fp, u'html.parser')

    if mode == u'standalone':
        _compose_for_standalone(osexp, dom, assets, params)
    elif mode == u'jatos':
        _compose_for_jatos(osexp, dom, assets, params)

    html = dom.prettify()
    with io.open(dst, 'w', encoding=u'utf-8') as fd:
        fd.write(safe_decode(html))

    return assets


def _compose_for_standalone(osexp, dom, assets, params=None):
    """ Builds on top of the base HTML template to create a structure that is appropriate
    for a standalone HTML file """

    logo_tag = dom.new_tag(u'script', type=u'text/javascript')
    logo_path = os.path.join(srcPaths['img'], u'opensesame.png')
    logo_tag.append(
        u'const logoSrc = "data:image/png;base64,{}"'.format(_read_b64(logo_path)))
    dom.head.append(logo_tag)

    if params:
        params_tag = dom.new_tag(u'script', type=u"text/javascript")
        params_tag.append(
            u'const params = JSON.parse(\'{}\')\n'.format(json.dumps(params)))
        dom.head.append(params_tag)
    script_tag = dom.new_tag(u'script', type=u"text/javascript")
    for js_file in assets['js']:
        if os.path.splitext(js_file['src'])[1] == '.js':
            script_tag.append(_read(js_file['src']) + '\n')
    dom.head.append(script_tag)

    css_tag = dom.new_tag(u'style')
    for css_file in assets['css']:
        if os.path.splitext(css_file['src'])[1] == '.css':
            css_tag.append(_read(css_file['src']) + '\n')
    dom.head.append(css_tag)

    # Add experiment as base64 encoded string
    exp_tag = dom.new_tag(
        u'embed',
        id=u'osexp_src',
        src=u'data:application/gzip;base64,' + _read_b64(osexp),
        style=u'display:none'
    )
    dom.body.append(exp_tag)


def _compose_for_jatos(osexp, dom, assets, params=None):
    """ Builds on top of the base HTML template to create a structure that is appropriate
    for integration in JATOS """

    scriptTag = dom.new_tag('script', id="parameters", type="text/javascript")
    if params:
        # Create a script node that exposes the experiment's parameters
        scriptTag.append(
            u'const params = JSON.parse(\'{}\')'.format(json.dumps(params)))

    # Get the OpenSesame experiment file name, and add a JS variable to its location
    scriptTag.append(u'const osexpFile = "{}"'.format(os.path.basename(osexp)))
    dom.head.append(scriptTag)

    # Add script nodes referencing the sources of all other required javascript files.
    for js_file in assets['js']:
        if os.path.splitext(js_file['src'])[1] == '.js':
            scriptTag = dom.new_tag('script', src=js_file['dest'], type="text/javascript")
            dom.head.append(scriptTag)

    for css_file in assets['css']:
        if os.path.splitext(css_file['src'])[1] == '.css':
            styleTag = dom.new_tag('link', href=css_file['dest'], type="text/css", rel="stylesheet", media="all")
            dom.head.append(styleTag)


def _read(path):

    with io.open(path, encoding='utf-8') as fd:
        return fd.read()


def _read_b64(path):

    with open(path, 'rb') as fd:
        e = base64.b64encode(fd.read())
        return e.decode()


def _unique_hash():

    return hashlib.md5(str(time.time()).encode()).hexdigest()


if __name__ == "__main__":
    standalone('', 'output.html')
